import json


async def test_list_formats(jp_fetch):
    r = await jp_fetch(
        'api', 'nbconvert',
        method='GET'
    )
    formats = json.loads(r.body.decode())
    assert isinstance(formats, dict)
    assert 'python' in formats
    assert 'html' in formats
    assert formats['python']['output_mimetype'] == 'text/x-python'



from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin, ExtensionHandlerJinjaMixin
from jupyter_server.utils import url_escape

class ParameterHandler(ExtensionHandlerMixin, JupyterHandler):
    def get(self, matched_part=None, *args, **kwargs):
        var1 = self.get_argument('var1', default=None)
        components = [x for x in self.request.path.split("/") if x]
        self.write('<h1>Hello Simple App 2 from Handler.</h1>')
        self.write('<p>matched_part: {}</p>'.format(url_escape(matched_part)))
        self.write('<p>var1: {}</p>'.format(url_escape(var1)))
        self.write('<p>components: {}</p>'.format(components))

class BaseTemplateHandler(ExtensionHandlerJinjaMixin, ExtensionHandlerMixin, JupyterHandler): pass

class IndexHandler(BaseTemplateHandler):
    def get(self):
        self.write(self.render_template("index.html"))

class TemplateHandler(BaseTemplateHandler):
    def get(self, path):
        print(self.get_template('simple_ext2.html'))
        self.write(self.render_template('simple_ext2.html', path=path))

class ErrorHandler(BaseTemplateHandler):
    def get(self, path):
        self.write(self.render_template('error.html'))



import os
from setuptools import setup
from jupyter_packaging import create_cmdclass


VERSION = '0.0.1'


def get_data_files():
    """Get the data files for the package.
    """
    data_files = [
        ('etc/jupyter/jupyter_server_config.d', 'etc/jupyter/jupyter_server_config.d/', '*.json'),
    ]
    def add_data_files(path):
        for (dirpath, dirnames, filenames) in os.walk(path):
            if filenames:
                paths = [(dirpath, dirpath, filename) for filename in filenames]
                data_files.extend(paths)
    # Add all static and templates folders.
    add_data_files('simple_ext1/static')
    add_data_files('simple_ext1/templates')
    add_data_files('simple_ext2/static')
    add_data_files('simple_ext2/templates')
    return data_files


cmdclass = create_cmdclass(
    data_files_spec=get_data_files()
)

setup_args = dict(
    name = 'jupyter_server_example',
    version = VERSION,
    description = 'Jupyter Server Example',
    long_description = open('README.md').read(),
    python_requires = '>=3.6',
    install_requires = [
        'jupyter_server',
        'jinja2',
    ],
    extras_require = {
        'test': ['pytest'],
    },
    include_package_data=True,
    cmdclass = cmdclass,
    entry_points = {
        'console_scripts': [
             'jupyter-simple-ext1 = simple_ext1.application:main',
             'jupyter-simple-ext11 = simple_ext11.application:main',
             'jupyter-simple-ext2 = simple_ext2.application:main'
        ]
    },
)


if __name__ == '__main__':
    setup(**setup_args)



try:
    from anyio.to_thread import run_sync
except ImportError:
    # fallback on anyio v2 for python version < 3.7
    from anyio import run_sync_in_worker_thread as run_sync
    
from tornado import web
import base64
import os, io

from jupyter_server.services.contents.filemanager import AsyncFileContentsManager, FileContentsManager


class LargeFileManager(FileContentsManager):
    """Handle large file upload."""

    def save(self, model, path=''):
        """Save the file model and return the model with no content."""
        chunk = model.get('chunk', None)
        if chunk is not None:
            path = path.strip('/')

            if 'type' not in model:
                raise web.HTTPError(400, u'No file type provided')
            if model['type'] != 'file':
                raise web.HTTPError(400, u'File type "{}" is not supported for large file transfer'.format(model['type']))
            if 'content' not in model and model['type'] != 'directory':
                raise web.HTTPError(400, u'No file content provided')

            os_path = self._get_os_path(path)

            try:
                if chunk == 1:
                    self.log.debug("Saving %s", os_path)
                    self.run_pre_save_hook(model=model, path=path)
                    super(LargeFileManager, self)._save_file(os_path, model['content'], model.get('format'))
                else:
                    self._save_large_file(os_path, model['content'], model.get('format'))
            except web.HTTPError:
                raise
            except Exception as e:
                self.log.error(u'Error while saving file: %s %s', path, e, exc_info=True)
                raise web.HTTPError(500, u'Unexpected error while saving file: %s %s' %
                                    (path, e)) from e

            model = self.get(path, content=False)

            # Last chunk
            if chunk == -1:
                self.run_post_save_hook(model=model, os_path=os_path)
            return model
        else:
            return super(LargeFileManager, self).save(model, path)

    def _save_large_file(self, os_path, content, format):
        """Save content of a generic file."""
        if format not in {'text', 'base64'}:
            raise web.HTTPError(
                400,
                "Must specify format of file contents as 'text' or 'base64'",
            )
        try:
            if format == 'text':
                bcontent = content.encode('utf8')
            else:
                b64_bytes = content.encode('ascii')
                bcontent = base64.b64decode(b64_bytes)
        except Exception as e:
            raise web.HTTPError(
                400, u'Encoding error saving %s: %s' % (os_path, e)
            ) from e

        with self.perm_to_403(os_path):
            if os.path.islink(os_path):
                os_path = os.path.join(os.path.dirname(os_path), os.readlink(os_path))
            with io.open(os_path, 'ab') as f:
                f.write(bcontent)


class AsyncLargeFileManager(AsyncFileContentsManager):
    """Handle large file upload asynchronously"""

    async def save(self, model, path=''):
        """Save the file model and return the model with no content."""
        chunk = model.get('chunk', None)
        if chunk is not None:
            path = path.strip('/')

            if 'type' not in model:
                raise web.HTTPError(400, u'No file type provided')
            if model['type'] != 'file':
                raise web.HTTPError(400, u'File type "{}" is not supported for large file transfer'.format(model['type']))
            if 'content' not in model and model['type'] != 'directory':
                raise web.HTTPError(400, u'No file content provided')

            os_path = self._get_os_path(path)

            try:
                if chunk == 1:
                    self.log.debug("Saving %s", os_path)
                    self.run_pre_save_hook(model=model, path=path)
                    await super(AsyncLargeFileManager, self)._save_file(os_path, model['content'], model.get('format'))
                else:
                    await self._save_large_file(os_path, model['content'], model.get('format'))
            except web.HTTPError:
                raise
            except Exception as e:
                self.log.error(u'Error while saving file: %s %s', path, e, exc_info=True)
                raise web.HTTPError(500, u'Unexpected error while saving file: %s %s' %
                                    (path, e)) from e

            model = await self.get(path, content=False)

            # Last chunk
            if chunk == -1:
                self.run_post_save_hook(model=model, os_path=os_path)
            return model
        else:
            return await super(AsyncLargeFileManager, self).save(model, path)

    async def _save_large_file(self, os_path, content, format):
        """Save content of a generic file."""
        if format not in {'text', 'base64'}:
            raise web.HTTPError(
                400,
                "Must specify format of file contents as 'text' or 'base64'",
            )
        try:
            if format == 'text':
                bcontent = content.encode('utf8')
            else:
                b64_bytes = content.encode('ascii')
                bcontent = base64.b64decode(b64_bytes)
        except Exception as e:
            raise web.HTTPError(
                400, u'Encoding error saving %s: %s' % (os_path, e)
            ) from e

        with self.perm_to_403(os_path):
            if os.path.islink(os_path):
                os_path = os.path.join(os.path.dirname(os_path), os.readlink(os_path))
            with io.open(os_path, 'ab') as f:
                await run_sync(f.write, bcontent)





"""
store the current version info of the server.

"""
from jupyter_packaging import get_version_info

# Version string must appear intact for tbump versioning
__version__ = '1.7.0a1'
version_info = get_version_info(__version__)



"""A mock extension module with a list of extensions
to load in various tests.
"""
from .app import MockExtensionApp


# Function that makes these extensions discoverable
# by the test functions.
def _jupyter_server_extension_points():
    return [
        {
            'module': 'jupyter_server.tests.extension.mockextensions.app',
            'app': MockExtensionApp
        },
        {
            'module': 'jupyter_server.tests.extension.mockextensions.mock1'
        },
        {
            'module': 'jupyter_server.tests.extension.mockextensions.mock2'
        },
        {
            'module': 'jupyter_server.tests.extension.mockextensions.mock3'
        }
    ]



"""Test launching Jupyter Server Applications
through as ExtensionApp launch_instance.
"""
from pathlib import Path
import os
import sys
import time
import pytest
import subprocess
import requests
from binascii import hexlify


HERE = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def port():
    return 9999


@pytest.fixture
def token():
    return hexlify(os.urandom(4)).decode("ascii")


@pytest.fixture
def auth_header(token):
    return {
        'Authorization': 'token %s' % token
    }


def wait_up(url, interval=0.1, check=None):
    while True:
        try:
            r = requests.get(url)
        except Exception:
            if check:
                assert check()
            #print("waiting for %s" % url)
            time.sleep(interval)
        else:
            break


@pytest.fixture
def launch_instance(request, port, token):
    def _run_in_subprocess(argv=[], add_token=True):

        def _kill_extension_app():
            try:
                process.terminate()
            except OSError:
                # Already dead.
                pass
            process.wait(10)

        if add_token:
            f'--ServerApp.token="{token}"',

        process = subprocess.Popen([
            sys.executable, '-m',
            'mockextensions.app',
            f'--port={port}',
            '--ip=127.0.0.1',
            '--no-browser',
            *argv,
        ], cwd=HERE)

        request.addfinalizer(_kill_extension_app)
        url = f'http://127.0.0.1:{port}'
        wait_up(url, check=lambda: process.poll() is None)
        return process

    return _run_in_subprocess


@pytest.fixture
def fetch(port, auth_header):
    def _get(endpoint):
        url = f"http://127.0.0.1:{port}" + endpoint
        return requests.get(url, headers=auth_header)
    return _get


def test_launch_instance(launch_instance, fetch):
    launch_instance()
    r = fetch('/mock')
    assert r.status_code == 200


def test_base_url(launch_instance, fetch):
    launch_instance(['--ServerApp.base_url=/foo'])
    r = fetch("/foo/mock")
    assert r.status_code == 200


def test_token_file(launch_instance, fetch, token):
    token_file = HERE / Path('token_file.txt')
    os.environ['JUPYTER_TOKEN_FILE'] = str(token_file)
    token_file.write_text(token, encoding='utf-8')

    launch_instance(add_token=False)
    r = fetch("/mock")
    del os.environ['JUPYTER_TOKEN_FILE']
    token_file.unlink()
    assert r.status_code == 200





import pytest

from jupyter_core.paths import jupyter_config_path
from jupyter_server.extension.config import (
    ExtensionConfigManager,
)

# Use ServerApps environment because it monkeypatches
# jupyter_core.paths and provides a config directory
# that's not cross contaminating the user config directory.
pytestmark = pytest.mark.usefixtures("jp_environ")


@pytest.fixture
def configd(jp_env_config_path):
    """A pathlib.Path object that acts like a jupyter_server_config.d folder."""
    configd = jp_env_config_path.joinpath('jupyter_server_config.d')
    configd.mkdir()
    return configd


ext1_json_config = """\
{
    "ServerApp": {
        "jpserver_extensions": {
            "ext1_config": true
        }
    }
}
"""

@pytest.fixture
def ext1_config(configd):
    config = configd.joinpath("ext1_config.json")
    config.write_text(ext1_json_config)


ext2_json_config = """\
{
    "ServerApp": {
        "jpserver_extensions": {
            "ext2_config": false
        }
    }
}
"""


@pytest.fixture
def ext2_config(configd):
    config = configd.joinpath("ext2_config.json")
    config.write_text(ext2_json_config)


def test_list_extension_from_configd(ext1_config, ext2_config):
    manager = ExtensionConfigManager(
        read_config_path=jupyter_config_path()
    )
    extensions = manager.get_jpserver_extensions()
    assert "ext2_config" in extensions
    assert "ext1_config" in extensions


from setuptools import setup

try:
    from jupyter_packaging import wrap_installers, npm_builder
    ensured_targets = ['jupyter_server/static/style/bootstrap.min.css']
    cmdclass = wrap_installers(pre_develop=npm_builder(), ensured_targets=ensured_targets)
except ImportError:
    cmdclass = {}

setup(cmdclass=cmdclass)



import os

import pytest

from jupyter_core.paths import jupyter_config_path

from jupyter_server.extension.manager import (
    ExtensionPoint,
    ExtensionPackage,
    ExtensionManager,
    ExtensionMetadataError,
    ExtensionModuleNotFound
)

# Use ServerApps environment because it monkeypatches
# jupyter_core.paths and provides a config directory
# that's not cross contaminating the user config directory.
pytestmark = pytest.mark.usefixtures("jp_environ")


def test_extension_point_api():
    # Import mock extension metadata
    from .mockextensions import _jupyter_server_extension_points

    # Testing the first path (which is an extension app).
    metadata_list = _jupyter_server_extension_points()
    point = metadata_list[0]

    module = point["module"]
    app = point["app"]

    e = ExtensionPoint(metadata=point)
    assert e.module_name == module
    assert e.name == app.name
    assert app is not None
    assert callable(e.load)
    assert callable(e.link)
    assert e.validate()


def test_extension_point_metadata_error():
    # Missing the "module" key.
    bad_metadata = {"name": "nonexistent"}
    with pytest.raises(ExtensionMetadataError):
        ExtensionPoint(metadata=bad_metadata)


def test_extension_point_notfound_error():
    bad_metadata = {"module": "nonexistent"}
    with pytest.raises(ExtensionModuleNotFound):
        ExtensionPoint(metadata=bad_metadata)


def test_extension_package_api():
    # Import mock extension metadata
    from .mockextensions import _jupyter_server_extension_points

    # Testing the first path (which is an extension app).
    metadata_list = _jupyter_server_extension_points()
    path1 = metadata_list[0]
    app = path1["app"]

    e = ExtensionPackage(name='jupyter_server.tests.extension.mockextensions')
    e.extension_points
    assert hasattr(e, "extension_points")
    assert len(e.extension_points) == len(metadata_list)
    assert app.name in e.extension_points
    assert e.validate()


def test_extension_package_notfound_error():
    with pytest.raises(ExtensionModuleNotFound):
        ExtensionPackage(name="nonexistent")


def _normalize_path(path_list):
    return [p.rstrip(os.path.sep) for p in path_list]


def test_extension_manager_api():
    jpserver_extensions = {
        "jupyter_server.tests.extension.mockextensions": True
    }
    manager = ExtensionManager()
    assert manager.config_manager
    assert _normalize_path(manager.config_manager.read_config_path) == _normalize_path(jupyter_config_path())
    manager.from_jpserver_extensions(jpserver_extensions)
    assert len(manager.extensions) == 1
    assert "jupyter_server.tests.extension.mockextensions" in manager.extensions


def test_extension_manager_linked_extensions(jp_serverapp):
    name = "jupyter_server.tests.extension.mockextensions"
    manager = ExtensionManager()
    manager.add_extension(name, enabled=True)
    manager.link_extension(name, jp_serverapp)
    assert name in manager.linked_extensions



import os
from simple_ext1.application import SimpleApp1
from jupyter_server.serverapp import aliases, flags
from traitlets import Bool, Unicode, observe

DEFAULT_STATIC_FILES_PATH = os.path.join(os.path.dirname(__file__), "./../simple_ext1/static")
DEFAULT_TEMPLATE_FILES_PATH = os.path.join(os.path.dirname(__file__), "./../simple_ext1/templates")

class SimpleApp11(SimpleApp1):
    flags['hello']=(
        { 'SimpleApp11' : {'hello' : True} }, "Say hello on startup."
    )
    aliases.update({
        'simple11-dir': 'SimpleApp11.simple11_dir',
    })

    # The name of the extension.
    name = "simple_ext11"

    # Te url that your extension will serve its homepage.
    extension_url = '/simple_ext11/default'

    # Local path to static files directory.
    static_paths = [
        DEFAULT_STATIC_FILES_PATH
    ]

    # Local path to templates directory.
    template_paths = [
        DEFAULT_TEMPLATE_FILES_PATH
    ]

    simple11_dir = Unicode('',
        config=True,
        help='Simple directory'
    )

    hello = Bool(False,
        config=True,
        help='Say hello',
    )

    ignore_js = Bool(False,
        config=True,
        help='Ignore Javascript',
    )

    @observe('ignore_js')
    def _update_ignore_js(self, change):
        """TODO Does the observe work?"""
        self.log.info('ignore_js has just changed: {}'.format(change))

    @property
    def simple11_dir_formatted(self):
        return "/" + self.simple11_dir

    def initialize_settings(self):
        self.log.info('hello: {}'.format(self.hello))
        if self.hello == True:
            self.log.info("Hello Simple11: You have launched with --hello flag or defined 'c.SimpleApp1.hello == True' in your config file")
        self.log.info('ignore_js: {}'.format(self.ignore_js))
        super().initialize_settings()

    def initialize_handlers(self):
        super().initialize_handlers()

#-----------------------------------------------------------------------------
# Main entry point
#-----------------------------------------------------------------------------

main = launch_new_instance = SimpleApp11.launch_instance
